@@ -66,6 +66,7 @@ import org.slf4j.LoggerFactory;
 import java.io.IOException;
 import java.net.InetSocketAddress;
 import java.nio.ByteBuffer;
+import java.util.Collections;
 import java.util.List;
 import java.util.Optional;
 import java.util.concurrent.CompletableFuture;
@@ -142,7 +143,7 @@ public class NettyClientStreamRpc implements DataStreamClientRpc {
     private final Supplier<ChannelInitializer<SocketChannel>> channelInitializerSupplier;
 
     /** The {@link ChannelFuture} is null when this connection is closed. */
-    private final AtomicReference<Supplier<ChannelFuture>> ref;
+    private final AtomicReference<MemoizedSupplier<ChannelFuture>> ref;
 
     Connection(InetSocketAddress address, WorkerGroupGetter workerGroup,
         Supplier<ChannelInitializer<SocketChannel>> channelInitializerSupplier) {
@@ -175,6 +176,9 @@ public class NettyClientStreamRpc implements DataStreamClientRpc {
     }
 
     private ChannelFuture connect() {
+      if (isClosed()) {
+        return null;
+      }
       return new Bootstrap()
           .group(getWorkerGroup())
           .channel(NettyUtils.getSocketChannelClass(getWorkerGroup()))
@@ -185,15 +189,18 @@ public class NettyClientStreamRpc implements DataStreamClientRpc {
             @Override
             public void operationComplete(ChannelFuture future) {
               if (!future.isSuccess()) {
-                scheduleReconnect(this + " failed", future.cause());
+                scheduleReconnect(Connection.this + " failed", future.cause());
               } else {
-                LOG.trace("{} succeed.", this);
+                LOG.trace("{} succeed.", Connection.this);
               }
             }
           });
     }
 
     void scheduleReconnect(String message, Throwable cause) {
+      if (isClosed()) {
+        return;
+      }
       LOG.warn("{}: {}; schedule reconnecting to {} in {}", this, message, address, RECONNECT);
       if (cause != null) {
         LOG.warn("", cause);
@@ -215,20 +222,22 @@ public class NettyClientStreamRpc implements DataStreamClientRpc {
       // AtomicReference.getAndUpdate may call the update function multiple times and discard the old objects.
       // The outer supplier creates only an inner supplier, which can be discarded without any leakage.
       // The inner supplier will be invoked (i.e. connect) ONLY IF it is successfully set to the reference.
-      final MemoizedSupplier<Supplier<ChannelFuture>> supplier = MemoizedSupplier.valueOf(
+      final MemoizedSupplier<MemoizedSupplier<ChannelFuture>> supplier = MemoizedSupplier.valueOf(
           () -> MemoizedSupplier.valueOf(this::connect));
-      final Supplier<ChannelFuture> previous = ref.getAndUpdate(prev -> prev == null? null: supplier.get());
-      if (previous != null) {
+      final MemoizedSupplier<ChannelFuture> previous = ref.getAndUpdate(prev -> prev == null? null: supplier.get());
+      if (previous != null && previous.isInitialized()) {
         previous.get().channel().close();
       }
       return getChannelFuture();
     }
 
     void close() {
-      final Supplier<ChannelFuture> previous = ref.getAndSet(null);
-      if (previous != null) {
+      final MemoizedSupplier<ChannelFuture> previous = ref.getAndSet(null);
+      if (previous != null && previous.isInitialized()) {
         // wait channel closed, do shutdown workerGroup
         previous.get().channel().close().addListener(future -> workerGroup.shutdownGracefully());
+      } else {
+        workerGroup.shutdownGracefully();
       }
     }
 
@@ -242,29 +251,44 @@ public class NettyClientStreamRpc implements DataStreamClientRpc {
     }
   }
 
-  class OutstandingRequests {
+  static class OutstandingRequests {
     private int count;
     private long bytes;
 
-    synchronized boolean write(DataStreamRequest request) {
-      count++;
-      bytes += request.getDataLength();
-      final List<WriteOption> options = request.getWriteOptionList();
-      final boolean isClose = options.contains(StandardWriteOption.CLOSE);
-      final boolean isFlush = options.contains(StandardWriteOption.FLUSH);
-      final boolean flush = shouldFlush(isClose || isFlush, flushRequestCountMin, flushRequestBytesMin);
-      LOG.debug("Stream{} outstanding: count={}, bytes={}, options={}, flush? {}",
-          request.getStreamId(), count, bytes, options, flush);
-      return flush;
+    private boolean shouldFlush(List<WriteOption> options, int countMin, SizeInBytes bytesMin) {
+      if (options.contains(StandardWriteOption.CLOSE)) {
+        // flush in order to send the CLOSE option.
+        return true;
+      } else if (bytes == 0 && count == 0) {
+        // nothing to flush (when bytes == 0 && count > 0, client may have written empty packets for including options)
+        return false;
+      } else {
+        return count >= countMin
+            || bytes >= bytesMin.getSize()
+            || options.contains(StandardWriteOption.FLUSH);
+      }
     }
 
-    synchronized boolean shouldFlush(boolean force, int countMin, SizeInBytes bytesMin) {
-      if (force || count >= countMin || bytes >= bytesMin.getSize()) {
+    synchronized boolean shouldFlush(int countMin, SizeInBytes bytesMin, DataStreamRequest request) {
+      final List<WriteOption> options;
+      if (request == null) {
+        options = Collections.emptyList();
+      } else {
+        options = request.getWriteOptionList();
+        count++;
+        final long length = request.getDataLength();
+        Preconditions.assertTrue(length >= 0, () -> "length = " + length + " < 0, request: " + request);
+        bytes += length;
+      }
+
+      final boolean flush = shouldFlush(options, countMin, bytesMin);
+      LOG.debug("flush? {}, (count, bytes)=({}, {}), min=({}, {}), request={}, options={}",
+          flush, count, bytes, countMin, bytesMin, request, options);
+      if (flush) {
         count = 0;
         bytes = 0;
-        return true;
       }
-      return false;
+      return flush;
     }
   }
 
@@ -280,7 +304,7 @@ public class NettyClientStreamRpc implements DataStreamClientRpc {
   private final OutstandingRequests outstandingRequests = new OutstandingRequests();
 
   public NettyClientStreamRpc(RaftPeer server, TlsConf tlsConf, RaftProperties properties) {
-    this.name = JavaUtils.getClassSimpleName(getClass()) + "->" + server;
+    this.name = JavaUtils.getClassSimpleName(getClass()) + "->" + server.getId();
     this.requestTimeout = RaftClientConfigKeys.DataStream.requestTimeout(properties);
     this.closeTimeout = requestTimeout.multiply(2);
 
@@ -299,16 +323,16 @@ public class NettyClientStreamRpc implements DataStreamClientRpc {
       @Override
       public void channelRead(ChannelHandlerContext ctx, Object msg) {
         if (!(msg instanceof DataStreamReply)) {
-          LOG.error("{}: unexpected message {}", this, msg.getClass());
+          LOG.error("{}: unexpected message {}", name, msg.getClass());
           return;
         }
         final DataStreamReply reply = (DataStreamReply) msg;
-        LOG.debug("{}: read {}", this, reply);
+        LOG.debug("{}: read {}", name, reply);
         final ClientInvocationId clientInvocationId = ClientInvocationId.valueOf(
             reply.getClientId(), reply.getStreamId());
         final NettyClientReplies.ReplyMap replyMap = replies.getReplyMap(clientInvocationId);
         if (replyMap == null) {
-          LOG.error("{}: {} replyMap not found for reply: {}", this, clientInvocationId, reply);
+          LOG.error("{}: {} replyMap not found for reply: {}", name, clientInvocationId, reply);
           return;
         }
 
@@ -329,9 +353,7 @@ public class NettyClientStreamRpc implements DataStreamClientRpc {
 
       @Override
       public void channelInactive(ChannelHandlerContext ctx) {
-        if (!connection.isClosed()) {
-          connection.scheduleReconnect("channel is inactive", null);
-        }
+        connection.scheduleReconnect("channel is inactive", null);
       }
     };
   }
@@ -417,8 +439,8 @@ public class NettyClientStreamRpc implements DataStreamClientRpc {
         return f;
       }
       replyEntry = replyMap.submitRequest(requestEntry, isClose, f);
-      final Function<DataStreamRequest, ChannelFuture> writeMethod = outstandingRequests.write(request)?
-          channel::writeAndFlush: channel::write;
+      final Function<DataStreamRequest, ChannelFuture> writeMethod = outstandingRequests.shouldFlush(
+          flushRequestCountMin, flushRequestBytesMin, request)? channel::writeAndFlush: channel::write;
       channelFuture = writeMethod.apply(request);
     }
     channelFuture.addListener(future -> {
@@ -448,8 +470,7 @@ public class NettyClientStreamRpc implements DataStreamClientRpc {
 
   @Override
   public void close() {
-    final boolean flush = outstandingRequests.shouldFlush(true, 0, SizeInBytes.ZERO);
-    LOG.debug("flush? {}", flush);
+    final boolean flush = outstandingRequests.shouldFlush(0, SizeInBytes.ZERO, null);
     if (flush) {
       Optional.ofNullable(connection.getChannelUninterruptibly())
           .map(c -> c.writeAndFlush(EMPTY_BYTE_BUFFER))
