@@ -59,6 +59,7 @@ import org.apache.ratis.util.Preconditions;
 import org.apache.ratis.util.ReferenceCountedObject;
 import org.apache.ratis.util.TimeDuration;
 import org.apache.ratis.util.TimeoutExecutor;
+import org.apache.ratis.util.UncheckedAutoCloseable;
 import org.apache.ratis.util.function.CheckedBiFunction;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
@@ -76,9 +77,11 @@ import java.util.concurrent.CompletionException;
 import java.util.concurrent.ConcurrentHashMap;
 import java.util.concurrent.ConcurrentMap;
 import java.util.concurrent.Executor;
+import java.util.concurrent.ExecutorService;
 import java.util.concurrent.atomic.AtomicReference;
 import java.util.function.Function;
 import java.util.stream.Collectors;
+import java.util.stream.Stream;
 
 public class DataStreamManagement {
   public static final Logger LOG = LoggerFactory.getLogger(DataStreamManagement.class);
@@ -118,10 +121,18 @@ public class DataStreamManagement {
       this.out = out;
     }
 
+    static Iterable<WriteOption> addFlush(List<WriteOption> original) {
+      if (original.contains(StandardWriteOption.FLUSH)) {
+        return original;
+      }
+      return Stream.concat(Stream.of(StandardWriteOption.FLUSH), original.stream())
+          .collect(Collectors.toList());
+    }
+
     CompletableFuture<DataStreamReply> write(DataStreamRequestByteBuf request, Executor executor) {
       final Timekeeper.Context context = metrics.start();
       return composeAsync(sendFuture, executor,
-          n -> out.writeAsync(request.slice().nioBuffer(), request.getWriteOptionList())
+          n -> out.writeAsync(request.slice().nioBuffer(), addFlush(request.getWriteOptionList()))
               .whenComplete((l, e) -> metrics.stop(context, e == null)));
     }
   }
@@ -254,8 +265,8 @@ public class DataStreamManagement {
 
   private final StreamMap streams = new StreamMap();
   private final ChannelMap channels;
-  private final Executor requestExecutor;
-  private final Executor writeExecutor;
+  private final ExecutorService requestExecutor;
+  private final ExecutorService writeExecutor;
 
   private final NettyServerStreamRpcMetrics nettyServerStreamRpcMetrics;
 
@@ -276,6 +287,13 @@ public class DataStreamManagement {
     this.nettyServerStreamRpcMetrics = metrics;
   }
 
+  void shutdown() {
+    ConcurrentUtils.shutdownAndWait(TimeDuration.ONE_SECOND, requestExecutor,
+        timeout -> LOG.warn("{}: requestExecutor shutdown timeout in {}", this, timeout));
+    ConcurrentUtils.shutdownAndWait(TimeDuration.ONE_SECOND, writeExecutor,
+        timeout -> LOG.warn("{}: writeExecutor shutdown timeout in {}", this, timeout));
+  }
+
   private CompletableFuture<DataStream> stream(RaftClientRequest request, StateMachine stateMachine) {
     final RequestMetrics metrics = getMetrics().newRequestMetrics(RequestType.STATE_MACHINE_STREAM);
     final Timekeeper.Context context = metrics.start();
@@ -327,14 +345,12 @@ public class DataStreamManagement {
     final DataChannel channel = stream.getDataChannel();
     long byteWritten = 0;
     for (ByteBuffer buffer : buf.nioBuffers()) {
-      final ReferenceCountedObject<ByteBuffer> wrapped = ReferenceCountedObject.wrap(buffer, buf::retain, buf::release);
-      wrapped.retain();
-      try {
+      final ReferenceCountedObject<ByteBuffer> wrapped = ReferenceCountedObject.wrap(
+          buffer, buf::retain, ignored -> buf.release());
+      try(UncheckedAutoCloseable ignore = wrapped.retainAndReleaseOnClose()) {
         byteWritten += channel.write(wrapped);
       } catch (Throwable t) {
         throw new CompletionException(t);
-      } finally {
-        wrapped.release();
       }
     }
 
